import 'dart:typed_data';
import 'dart:math' as math;

/// Computes sample arrays to be used with examples.
class SampleData {

  /// Computes the sine function
  ///   f(x) = offset + ampl*sin(x+phase)
  /// Returns 2 arrays of length [nxvalues].
  ///  The first array contains the "x" values where f(x) is computed.
  ///  The second array contains [nperiods] (2*pi) of a sine wave.
  static List<Float64List> genSine(double ampl, double offset, double phase,
      int nperiods, int nxvalues) {
    double xmax = 2 * math.pi * nperiods;
    Float64List xValues = Float64List(nxvalues);
    Float64List yValues = Float64List(nxvalues);
    double x, y;

    for (int i = 0; i < nxvalues; i++) {
      x = (i * xmax) / nxvalues;
      y = ampl * math.sin(x + phase);
      xValues[i] = x;
      yValues[i] = y + offset;
    }

    return [xValues, yValues];
  }

  /// Computes the "sinc" function
  ///   f(x) = offset + ampl*sin(x+phase)/(x+phase)
  /// Returns 2 arrays of length [nxvalues].
  ///  The first array contains the "x" values where f(x) is computed.
  ///  The second array contains [nperiods] (2*pi) of a the damped sinc wave.
  static List<Float64List> genSinc(double ampl, double offset, double phase,
      int nperiods, int nxvalues) {
    double xmax = 2 * math.pi * nperiods;

    Float64List xValues = Float64List(nxvalues);
    Float64List yValues = Float64List(nxvalues);
    double x, y, a;

    a = ampl;

    for (int i = 0; i < nxvalues; i++) {
      x = (i * xmax) / nxvalues;
      if (i == 0 && phase.abs() < 0.00001) {
        y = a;
      } else {
        double ph = x + phase;
        y = a * math.sin(ph) / ph;
      }
      xValues[i] = x;
      yValues[i] = y + offset;
    }
    return [xValues, yValues];
  }
}
