import 'dart:convert';
import 'dart:html';
import 'dart:svg';
import 'dart:typed_data';
import 'dart:math' as math;

import '../smart_arrays_plot_polyline.dart';

/// This class allows you to display one or several arrays as
/// polylines (curves) inside the browser, along with coordinate axes and a
/// a coordinate grid, an some title text.
/// It is recommended to use this class if a plot needs to be generated with a few
/// lines of code, rather than exploitimg the full capabilities of the classes
/// [Polyline], [Axis], [Legend] for more complex layouts.
class SimplePlot {
  /// [plotDiv] will contain the graphics, it's size must be as wished by
  /// the user.
  DivElement plotDiv;

  /// All graphics is drawn using SVG Scalable Vector Graphics. [plotArea] is
  /// added to [plotDiv] and contains the entire plot. [dataArea] is added
  /// to [plotArea] and contains the polylines and the legend. The axes are
  /// drawn outside the [dataArea] but within the [plotArea].
  SvgSvgElement plotArea, dataArea;

  /// A frame around the 'dataArea'.
  RectElement dataAreaBorder;

  /// The coordinates with respect to their container
  math.Rectangle<int> plotAreaRect,
      dataAreaRect,
      xaxisRect,
      yaxisRect,
      dataInsets;

  /// The axes
  Axis xaxis, yaxis;

  /// The arrays to be plotted
  List<Float64List> arrays;

  List<Polyline> polylines;
  Legend legend;
  List<Map<PyA, String>> plotAttr;
  Map<LegA, String> legendAttr;
  Map<AxA, String> xaxisAttr, yaxisAttr;

  final List<String> POLY_COLORS = [
    "blue",
    "red",
    "magenta",
    "green",
    "orange",
    "cyan",
    "lightgreen",
    "crimson",
    "darkred",
    "darkgreen"
  ];

  List<String> poly_colors;

  /// Plots [arrays] as polylines (curves), along with coordinate axes,
  /// a coordinate grid and legend text.
  /// [plotDiv] - the html DivElement that will contain the plot.
  /// [plotAttr] - plot attributes controlling the appearance and layout of the
  ///   polylines. If [null], internal defaults are used.
  /// [xaxisAttr] - x axis attributes controlling the appearance and layout of the
  ///   x (horizontal) coordinate axis. If [null], internal defaults are used.
  /// [yaxisAttr] - y axis attributes controlling the appearance and layout of the
  ///   y (vertical) coordinate axis. If [null], internal defaults are used.
  /// [legendAttr] - attributes controlling the appearance and layout of the
  ///   text legend. If [null], internal defaults are used.
  /// [poly_colors] - these are the color in which [arrays] get displayed. This
  ///   list may be bigger than the list a [arrays], but not smaller.
  ///   If [null], internal default colors are used.
  /// See the example provided with this package how simple is it is to generate
  /// the plot.
  /// NOTE 1: Displaying more than a single array only makes sense if the value
  ///   ranges in the different arrays are commensurable, otherwise a simultaneous
  ///   plot is not meaningful.
  /// NOTE 2: The sizes of the arrays may be different. The width of the plot
  ///   is computed from [plotDiv].
  SimplePlot(this.arrays, this.plotDiv, this.plotAttr, this.xaxisAttr,
      this.yaxisAttr, this.legendAttr, this.poly_colors) {
    // Set up the containers where the graphics is to be drawn
    computeLayout();
    createPolylines();

    // Compute x and y axis for the first polyline
    createAxes(polylines[0]);

    createLegend();

    // append  polylines the "data area" of the layout
    for (int i = 0; i < arrays.length; i++) {
      dataArea.append(polylines[i].polylineContainer);
    }

    if (legendAttr != null) {
      // append  the legend to the "data area" of the layout
      dataArea.append(legend.legendContainer);
    }

    // append x axis, y axis and axes grid to the "plot area"  of the layout
    plotArea.append(xaxis.labelsContainer);
    plotArea.append(yaxis.labelsContainer);
    plotArea.append(xaxis.xyGrid.gridContainer);
    plotArea.append(yaxis.xyGrid.gridContainer);
  }

  /// Computes the plot layout: Sets up the graphics containers and their
  /// relative coordinates and sizes.
  void computeLayout() {
    // get divs from 'example.html'
//    plotDiv = (querySelector("#plot_div") as DivElement);

    // the plot takes all available space
//    plotDiv.style
//      ..width = "${appDiv.clientWidth}px"
//      ..height = "${appDiv.clientHeight}px";

    // Reserve space in pixels for the axes labels and a border around everthing.
    int xaxAreaHeight = 70, yaxAreaWidth = 80, borderAreaSize = 15;

    // will contain the entire plot
    plotAreaRect = math.Rectangle<int>(
        borderAreaSize,
        borderAreaSize,
        plotDiv.clientWidth - 2 * borderAreaSize,
        plotDiv.clientHeight - 2 * borderAreaSize);

    // all coordinates in the following rectangles are are relative to plotRect
    dataInsets = math.Rectangle<int>(borderAreaSize, borderAreaSize, 0, 0);

    // will contain the polylines and the legend
    dataAreaRect = math.Rectangle<int>(
        yaxAreaWidth,
        dataInsets.top,
        plotAreaRect.width - yaxAreaWidth - dataInsets.left,
        plotAreaRect.height - xaxAreaHeight);

    // will contain the x axis tic marks and text labels
    xaxisRect = math.Rectangle<int>(
        dataAreaRect.left,
        dataAreaRect.top + dataAreaRect.height,
        dataAreaRect.width,
        xaxAreaHeight);

    // will contain the y axis tic marks and text labels
    yaxisRect = math.Rectangle<int>(dataAreaRect.left - yaxAreaWidth,
        dataAreaRect.top, yaxAreaWidth, dataAreaRect.height);

    // will contain all axes and the data area (polylines and legend)
    plotArea = SvgSvgElement();
    SVG.setAttr(plotArea, {
      SVG.WIDTH: "${plotAreaRect.width}",
      SVG.HEIGHT: "${plotAreaRect.height}",
    });

    // will contain polylines and legend
    dataArea = SvgSvgElement();
    SVG.setAttr(dataArea, {
      SVG.X: "${dataAreaRect.left}",
      SVG.Y: "${dataAreaRect.top}",
      SVG.WIDTH: "${dataAreaRect.width}",
      SVG.HEIGHT: "${dataAreaRect.height}",
    });

    // a border around the data area
    dataAreaBorder = RectElement();
    SVG.setAttr(dataAreaBorder, {
      SVG.X: "${dataAreaRect.left}",
      SVG.Y: "${dataAreaRect.top}",
      SVG.WIDTH: "${dataAreaRect.width}",
      SVG.HEIGHT: "${dataAreaRect.height}",
      SVG.FILL: "none",
      SVG.STROKE: "darkgreen",
      SVG.STROKE_WIDTH: "2"
    });

    // now stack the plot containers.
    plotDiv.append(plotArea);
    plotArea.append(dataArea);
    plotArea.append(dataAreaBorder); // draw above dataArea
  } // computeLayout()

  /// Creates a polyline for each element of [arrays]
  createPolylines() {
    polylines = List<Polyline>(arrays.length);
    if (poly_colors == null) {
      poly_colors = POLY_COLORS;
    }
    final double YPOS0 = 0.8;

    for (int i = 0; i < arrays.length; i++) {
      if (i == 0 && (plotAttr == null || plotAttr.length != arrays.length)) {
        plotAttr = List<Map<PyA, String>>(arrays.length);
      }

      if (plotAttr[i] == null) {
        plotAttr[i] = Map.from(POLYLINE_DEFAULT_ATTRIBUTES);
        plotAttr[i][PyA.YPOSITION_ZERO] = "$YPOS0";
        plotAttr[i][PyA.STROKE] = poly_colors[i];
      } else {
        if (plotAttr[i][PyA.YPOSITION_ZERO] == null) {
          plotAttr[i][PyA.YPOSITION_ZERO] = "$YPOS0";
        }
        if (plotAttr[i][PyA.STROKE] == null) {
          plotAttr[i][PyA.STROKE] = poly_colors[i];
        }
        Map<PyA, String> attr = Map.from(POLYLINE_DEFAULT_ATTRIBUTES);
        attr.addAll(plotAttr[i]); // merge with defaults
        plotAttr[i] = attr;
      }

      if (i > 0) {
        plotAttr[i][PyA.REFYMAX] = "${polylines[0].ymax}";
      }

      polylines[i] = Polyline(
          arrays[i],
          0,
          arrays[i].length - 1,
          dataAreaRect.width,
          dataAreaRect.height,
          null,
          null,
          null,
          plotAttr[i]);
    }
  } // [createPolylines]

  /// Create an x and  a y axis for [polyline].
  ///
  void createAxes(Polyline polyline) {
    if (xaxisAttr == null) {
      xaxisAttr = Map.from(Axis.AXIS_DEFAULT_ATTRIBUTES);
      xaxisAttr[AxA.LEGENDTEXT] = "Point number";
    } else {
      Map<AxA, String> attr = Map.from(Axis.AXIS_DEFAULT_ATTRIBUTES);
      attr.addAll(xaxisAttr); // merge with defaults
      xaxisAttr = attr;
    }

    if (yaxisAttr == null) {
      yaxisAttr = Map.from(Axis.AXIS_DEFAULT_ATTRIBUTES);
      yaxisAttr[AxA.LEGENDTEXT] = "Function value";
    } else {
      Map<AxA, String> attr = Map.from(Axis.AXIS_DEFAULT_ATTRIBUTES);
      attr.addAll(yaxisAttr); // merge with defaults
      yaxisAttr = attr;
    }

    // physical units: use point index as defaults
    double physStart = 0;
    double physWidth = (polyline.array.length - 1).toDouble();
    if (xaxisAttr[AxA.PHYS_X_START] != null &&
        xaxisAttr[AxA.PHYS_X_START].isNotEmpty &&
        xaxisAttr[AxA.PHYS_X_WIDTH] != null &&
        xaxisAttr[AxA.PHYS_X_WIDTH].isNotEmpty) {
      physStart = double.parse(xaxisAttr[AxA.PHYS_X_START]);
      physWidth = double.parse(xaxisAttr[AxA.PHYS_X_WIDTH]);
    }

    // compute x axis
    xaxis = Axis.coord(
        polyline.array.length, // # points in the polyline's data array
        polyline.xValues.first, // display the array from this index
        polyline.xValues.last, // until this index
        physStart,
        physWidth,
        true, // mode for converting point index to physical unit
        null, // no extra calibration factor of axis labels
        dataAreaRect.width, // axis length in pixels
        xaxisRect.height, // height of axis area (reserves space for the labels)
        dataAreaRect.height, // grid line length, needed if 'grid' chosen
        polyline.xphysToXscreen, // method converting physical to screen coord.
        xaxisAttr,
        null // no mouse or touch interaction in this example
        );

    // set the x axis container positions as computed by [computeLayout].
    SVG.setAttr(xaxis.labelsContainer, {
      SVG.X: "${xaxisRect.left - xaxis.extra_space_for_edge_labels_x}",
      SVG.Y: "${xaxisRect.top}"
    });

    // set the x axis grid container positions as computed by [computeLayout].
    SVG.setAttr(xaxis.xyGrid.gridContainer,
        {SVG.X: "${dataAreaRect.left}", SVG.Y: "${dataAreaRect.top}"});

    // compute y axis (intensity axis)
    double firstY = polyline.yScreenToYphys(0);
    double lastY = polyline.yScreenToYphys(dataAreaRect.height);
    if (firstY == null) firstY = polyline.ymin;
    if (lastY == null) lastY = polyline.ymax;
    yaxis = Axis.intens(
        firstY / polyline.yscale, // axis range, account for [PyA.YSCALE]
        lastY / polyline.yscale,
        null, // no rescaling of y axis
        null, // like above, no rescaling of y axis
        null, // would only relevant for reversed axis
        dataAreaRect.height, // axis length in pixels
        yaxisRect.width, // height of axis area (reserves space for the labels)
        dataAreaRect.width, // grid line length, needed if 'grid' chosen
        polyline.yphysToYscreen, // method converting physical to screen coord.
        yaxisAttr,
        null // no mouse or touch interaction in this example
        );

    // set the y axis container positions as computed by [computeLayout].
    SVG.setAttr(yaxis.labelsContainer, {
      SVG.X: "${yaxisRect.left}",
      SVG.Y: "${yaxisRect.top - yaxis.extra_space_for_edge_labels_y}"
    });

    // set the y axis grid container positions as computed by [computeLayout].
    SVG.setAttr(yaxis.xyGrid.gridContainer,
        {SVG.X: "${dataAreaRect.left}", SVG.Y: "${dataAreaRect.top}"});
  }

  /// Sets up the legend
  void createLegend() {
    Map<LegA, String> attr = Map.from(Legend.LEGEND_DEFAULT_ATTRIBUTES);
    if (legendAttr != null) {
      attr.addAll(legendAttr); // merge with defaults
      legendAttr = attr;
    } else {
      legendAttr = attr;
    }

    if (legendAttr[LegA.TOP_TITLE] == null) {
      return;
    }

    int nsubtitles = 0;
    List<String> subtitles;
    // setup the subtitle lines, use same color as for polylines
    if (legendAttr[LegA.SUB_TITLES] != null) {
      subtitles = json.decode(legendAttr[LegA.SUB_TITLES]).cast<String>();
      nsubtitles = subtitles.length;
    }

    legend = Legend(legendAttr, nsubtitles); // create the legend

    // draw subtitles only if nsubtitles > 0:
    // - cut color list to the number needed
    // - note that setText draws from bottom up
    List<String> colors = List.from(poly_colors.sublist(0, nsubtitles));
    for (int i = nsubtitles - 1; i >= 0; i--) {
      legend.setText("${i + 1}", subtitles[i], colors[i]);
    }
  }
}
